


from . import benchmarks_00_Geometry_peakmem
from . import benchmarks_00_Geometry_time
from . import benchmarks_01_Mesh2D_peakmem
from . import benchmarks_01_Mesh2D_time
from . import benchmarks_02_spectralfit_peakmem
from . import benchmarks_02_spectralfit_time
from . import benchmarks_03_solidangles
